/* This code ported to djgpp & otherwise modified by:
   Marp (marp@technologist.com) &
   Sevy (sevy@zophar.net)
   for use with MSP (Miscellaneous Sound Player)
   http://smsoft.zophar.net */

/****************************************************************************
** Demonstration of playing an Audio CD                                    **
**  by Steven H Don                                                        **
**                                                                         **
** For questions, feel free to e-mail me.                                  **
**                                                                         **
**    shd@earthling.net                                                    **
**    http://shd.cjb.net                                                   **
**                                                                         **
****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <dpmi.h>
#include <sys/movedata.h>
#include "cdplayer.h"

/* Removed a ton of stuff and put it in cdplayer.h, look there ;-)  -Sevy*/

/**************************************************************************
** Check whether MSCDEX is installed and controlling at least one player **
**************************************************************************/
unsigned char Check_MSCDEX ()
{
        __dpmi_regs Regs;

        Regs.x.ax = 0x1500; /* Function 0 - Get number of CDROM players installed */
        Regs.x.bx = 0;
        __dpmi_int(0x2F, &Regs);
        if (Regs.x.bx)
        {
                /* If BX <> 0 then at least one player has been installed */
                /* Store the number of CD-ROM players */
                Number_Of_Players = Regs.x.bx;
                /* Store the drive indication of the first player */
                Drive_Letter      = Regs.x.cx;
                /* Code added by Sevy & Marp to detect the version of MSCDEX running and show it as well */
                Regs.x.ax = 0x150c;
                __dpmi_int(0x2f, &Regs);
                printf("MSCDEX version %i.%i found.\n", HIBYTE(Regs.x.bx), LOBYTE(Regs.x.bx));
                /* End of code added for version detection by Sevy & Marp */
                /* Note: This program only controls the first CD-ROM player installed,
                         in theory it is possible to control more than one player,
                         however, it is unlikely that someone either *has* more than
                         one player or that a programmer wishes to play music through
                         more than one player */
                /* Return true */
                return (1);
        }
        else
        {
                /* If BX = 0 then no players have been installed */
                /* Return false */
                return (0);
        }
}

/**************************************************************************
** Converts RedBook sector notation to HSG sector number.                **
***************************************************************************
**  When calculating the amount of sectors to be played, it's nice to    **
**  have a start and end offset so we can just subtract them.            **
**  This expects a Red Book sector address in the form of                **
**    MMMMSSFF (minutes * 10000h + seconds * 100h + frames)              **
**  and returns the LSN (logical sector number)                          **
**************************************************************************/
unsigned long HSG (unsigned long RedBook)
{
        unsigned char Frames;
        unsigned char Seconds;
        unsigned short Minutes;

        Frames  = (RedBook & 0xFF);
        Seconds = (RedBook & 0xFF00) >> 8;
        Minutes =  RedBook >> 16;
        return (Frames + (long)Seconds * 75 + (long)Minutes * 60 * 75);
}

/**************************************************************************
** Call MSCDEX which in turn calls the CDROM driver.                     **
***************************************************************************
**  This expects a control structure for a parameter.                    **
**************************************************************************/
void Call_CDROM_Driver(void *DrivParms, int size)
{
        __dpmi_regs Regs;
        int io_seg, io_sel;

        io_seg = __dpmi_allocate_dos_memory((size + 15) >> 4, &io_sel);

        /* Set up for call to MSCDEX, which in turn calls the driver */
        Regs.x.ax = 0x1510; /* Function 10 - call CDROM driver */
        Regs.x.cx = Drive_Letter; /* Drive letter for first player */
        Regs.x.es = io_seg; /* Pointer to Parameterblock */
        Regs.x.bx = 0;
        dosmemput(DrivParms, size, io_seg << 4);
        __dpmi_int(0x2F, &Regs); /* Call MSCDEX */
        dosmemget(io_seg << 4, size, DrivParms);
        __dpmi_free_dos_memory(io_sel);
}

/**************************************************************************
** Get information regarding the current CD.                             **
***************************************************************************
**  That information consists of                                         **
**    - The number of the first track on the CD                          **
**    - The number of the last track on the CD                           **
**    - The total playing length of the CD                               **
**************************************************************************/
void GetAudioCDInfo ()
{
        struct CDInfoStructure AudioCDInfo;
        int tb_seg, tb_sel;

        tb_seg = __dpmi_allocate_dos_memory((sizeof(AudioCDInfo) + 15) >> 4, &tb_sel);

        Command.FNr      = 3; /* Function 3 - Direct Read */
        Command.PBOfs    = 0;
        Command.PBSeg    = tb_seg;
        Command.Length   = 17;

        AudioCDInfo.SubFNr = 0xA; /* SubFunction 10 - get audio CD info */
        dosmemput(&AudioCDInfo, sizeof(AudioCDInfo), tb_seg << 4);
        Call_CDROM_Driver(&Command, sizeof(Command)); /* Call CDROM Driver */
        dosmemget(tb_seg << 4, sizeof(AudioCDInfo), &AudioCDInfo);
        __dpmi_free_dos_memory(tb_sel);
        First_Track   = AudioCDInfo.First;   //Store first track number
        Last_Track    = AudioCDInfo.Last;    //Store last track number
        LeadOut_Track = AudioCDInfo.LeadOut; //Store end of CD
        StatusWord    = Command.SW;
}

/**************************************************************************
** Starts playback of a specific track on the CD.                        **
**************************************************************************/
void PlayTrack (unsigned char Track)
{
        int tb_seg, tb_sel;
        struct TrackInfoStructure AudioTrackInfo;
        struct PlayStructure      Play;

        unsigned long             StartOffSet;
        unsigned long             StopOffSet;
        unsigned long             SectorsToPlay;

        tb_seg = __dpmi_allocate_dos_memory((sizeof(AudioTrackInfo) + 15) >> 4, &tb_sel);

        Command.FNr      = 3; /* Function 3 - Direct Read */
        Command.PBOfs    = 0;
        Command.PBSeg    = tb_seg;
        Command.Length   = 17;

        AudioTrackInfo.SubFNr  = 0xB;   /* Function 11 - Get audio track info */
        AudioTrackInfo.TrackNr = Track; /* Number of the track */
        dosmemput(&AudioTrackInfo, sizeof(AudioTrackInfo), tb_seg << 4);
        Call_CDROM_Driver(&Command, sizeof(Command));
        dosmemget(tb_seg << 4, sizeof(AudioTrackInfo), &AudioTrackInfo);

        /* Check to see whether the track is indeed an audio track */
        if ((AudioTrackInfo.Status & 0x40) == 0x40)
        {
                /* If not, display error message and don't try to start playing */
                printf ("Can't play data track!\n");
	        __dpmi_free_dos_memory(tb_sel);
                return;
        }

        StartOffSet = AudioTrackInfo.Address; //Start of the track

        /* End of the track */
        if (Track == Last_Track)
        {
                /* If it is the last, track, the end of the track is the LeadOut */
                StopOffSet = LeadOut_Track;
        }
        else
        {
                /* Otherwise, the end of this track is the start of the following track */
                Command.FNr      = 3; /* Function 3 - Direct Read */
                Command.PBOfs    = 0;
                Command.PBSeg    = tb_seg;
                Command.Length   = 17;

                AudioTrackInfo.SubFNr  = 0xB; /* Function 11 - Get audio track info */
                AudioTrackInfo.TrackNr = Track + 1; /* Number of the next track */
	        dosmemput(&AudioTrackInfo, sizeof(AudioTrackInfo), tb_seg << 4);
                Call_CDROM_Driver(&Command, sizeof(Command));
	        dosmemget(tb_seg << 4, sizeof(AudioTrackInfo), &AudioTrackInfo);
	        __dpmi_free_dos_memory(tb_sel);
                StopOffSet = AudioTrackInfo.Address;
        }

        /* Calculate the amount of sectors to be played */
        SectorsToPlay = HSG (StopOffSet) - HSG (StartOffSet);

        /* All right! That set up the start and end offsets of the track. Now we
           start the actual playing. */

        Play.FNr     = 0x84; /* Function 84h - play audio track */
        Play.Mode    = 1; /* Select Red Book addressing mode */
        Play.Start   = StartOffSet; /* Start of track */
        Play.Sectors = SectorsToPlay; /* Amount of sectors to play */
        Play.Length  = 21;
        Call_CDROM_Driver(&Play, sizeof(Play)); /* Call MSCDEX */
        StatusWord = Play.SW; /* Return status word */
}

/**************************************************************************
** Stops audio playback.                                                 **
**************************************************************************/
void StopPlaying ()
{
        struct StopStructure Stop;

        Stop.FNr     = 0x85; /* Function 85h - stop playback */
        Stop.Length  = 5;
        Call_CDROM_Driver(&Stop, sizeof(Stop)); /* Call MSCDEX */
        StatusWord = Stop.SW; /* Return status word */
}

/**************************************************************************
** Check the statusword to see if any errors have occurred.              **
***************************************************************************
**  This should be called after any function call                        **
**  Returns the number of the error                                      **
**************************************************************************/
unsigned char CheckForErrors ()
{
        //Check to see whether there was an error
        if ((StatusWord & 0x8000) > 0)
        {
                switch (StatusWord & 0xFF)
                {
                        case  0 : printf ("Can't write.\n"); break;
                        case  1 : printf ("Unknown player.\n"); break;
                        case  2 : printf ("No CD in player.\n"); break;
                        case  3 : printf ("Unknown function.\n"); break;
                        case  4 : printf ("CRC Error.\n"); break;
                        case  5 : printf ("Error in parameterblock.\n"); break;
                        case  6 : printf ("Seek fault.\n"); break;
                        case  7 : printf ("Unknown mediatype.\n"); break;
                        case  8 : printf ("Sector not found.\n"); break;
                        case  9 : printf ("Reserved.\n"); break;
                        case 10 : printf ("Write fault.\n"); break;
                        case 11 : printf ("Read fault.\n"); break;
                        case 12 : printf ("General failure error.\n"); break;
                        case 13 : printf ("Reserved.\n"); break;
                        case 14 : printf ("Reserved.\n"); break;
                        case 15 : printf ("Invalid CD Change.\n"); break;
                }
        return (StatusWord & 0xFF);
        }
        else
        {
                return (0);
        }
}



