/*
** Nofrendo - NES emulator
** (c) 1999 Matthew Conte
**
** License agreement:
**
** This source code may be distributed in unmodified form to any medium,
** and may be used in any *freeware* project.
**
** It may not be sold, in full or in part, as part of any commercial or
** shareware sofware package without the express written permission of 
** the author, Matthew Conte (matt@conte.com).
**
** Modified versions of this code may not be publicly distributed
** without the submission to, and approval of the author (matt@conte.com).
** This includes distribution in source, object, or binary form.
**
** The author claims no responsibility for damage cause by this code; it
** is distributed "as-is"; in other words, use at your own risk.
**
** Any freeware application using this code must credit the author,
** Matthew Conte, in the application itself or the documentation.
**
** Modifications should be reported to matt@conte.com, to ensure the
** highest degree of quality and accuracy is always freely available.
**
**
** log.c
**
** Error logging functions
*/

#include <stdio.h>
#include <stdarg.h>
#include "types.h"
#include "log.h"

#ifdef OSD_LOG
#include "osd.h"
#endif

#if defined(OSD_LOG) && !defined(NOFRENDO_DEBUG)
#error NOFRENDO_DEBUG must be defined as well as OSD_LOG
#endif

/* Note that all of these functions should be empty if
** debugging is not enabled.
*/
#ifdef NOFRENDO_DEBUG
static FILE *log;
static const char *logname = "errorlog.txt";
#endif

int log_init(void)
{
#ifdef NOFRENDO_DEBUG
#ifdef OSD_LOG
   /* Initialize an OSD logging system */
   osd_loginit();
#endif /* OSD_LOG */
   log = fopen(logname, "wt");
   if (NULL == log)
      return (-1);
#endif /* NOFRENDO_DEBUG */
   return 0;
}

void log_shutdown(void)
{
#ifdef NOFRENDO_DEBUG
#ifdef OSD_LOG
   osd_logshutdown();
#endif /* OSD_LOG */
   fclose(log);
#endif /* NOFRENDO_DEBUG */
}

void log_print(const char *string
#if !defined(NOFRENDO_DEBUG) && defined(__GNUC__)
__attribute__((unused))
#endif
)
{
#ifdef NOFRENDO_DEBUG
#ifdef OSD_LOG
   osd_logprint(string);
#endif /* OSD_LOG */
   /* Log it to disk, as well */
   fputs(string, log);
#endif /* NOFRENDO_DEBUG */
}

void log_printf(const char *format
#if !defined(NOFRENDO_DEBUG) && defined(__GNUC__)
__attribute__((unused))
#endif
, ... )
{
#ifdef NOFRENDO_DEBUG
#ifdef OSD_LOG
   char buffer[1024 + 1];
#endif /* OSD_LOG */
   va_list arg;

   va_start(arg, format);

#ifdef OSD_LOG
   vsprintf(buffer, format, arg);
   osd_logprint(buffer);
#endif /* OSD_LOG */
   vfprintf(log, format, arg);
   va_end(arg);
#endif /* NOFRENDO_DEBUG */
}
