/*

MSP v0.51 Souce Code
You can compile this with djgpp (GCC).
Compile using the supplied MAKEFILE and GNU MAKE.

You'll need the following libraries for a complete build:
1.  Allegro
2.  LibAmp (for MP3's)
3.  MikAllegro (for MOD's, etc.)
4.  zlib (for compressed GYM files)

If you would like to compile MSP without support for MOD's, MP3's, and
compressed GYM's, define ANOREXIC in the MAKEFILE (ANOREXIC=1). You will
then only need Allegro. Thanks to Matthew Conte for adding this option.

-Sevy

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <conio.h>
#include <math.h>
#include <time.h>
#include <dpmi.h>
#include <sys/movedata.h>

#include <allegro.h>

#ifndef ANOREXIC
#include <mikmod.h>
#include <libamp.h>
#include <zlib.h>
#endif

#include "ym2151.h"
#include "ym2612.h"
#include "sn76496.h"
#include "cdplayer.h"

/* for NSFs */
#include "types.h"
#include "nsf.h"
#include "nsf_play.h"
#include "nes_apu.h"
#include "log.h"


int wantloop; /* Used for determining whether to loop the sound or not */
int createconfig; /* Used for determing whether you want MSP to create a config file if none is available */

typedef struct {
    int id;
    char songtitle[32];
    char gametitle[32];
    char gamepub[32];
    char dumpemu[32];
    char dumper[32];
    char comments[256];
    unsigned long loop;
    unsigned long compressed;
} GYMTAG;

#ifndef ANOREXIC
/* Structure for holding ID3 tag. */
typedef struct {
    char ID[3],
         title[30],
         artist[30],
         album[30],
         year[4],
         comment[30];
    unsigned char genre;
} MP3TAG __attribute__((packed));

/* Function returns string decribing genre number for ID3 tag. */
char *get_genre_string(unsigned char genre_num) {
    switch(genre_num) {
        /* Officially defined genres */
        case(0): return "Blues";
        case(1): return "Classic Rock";
        case(2): return "Country";
        case(3): return "Dance";
        case(4): return "Disco";
        case(5): return "Funk";
        case(6): return "Grunge";
        case(7): return "Hip-hop";
        case(8): return "Jazz";
        case(9): return "Metal";
        case(10): return "New Age";
        case(11): return "Oldies";
        case(12): return "Other";
        case(13): return "Pop";
        case(14): return "R&B";
        case(15): return "Rap";
        case(16): return "Reggae";
        case(17): return "Rock";
        case(18): return "Techno";
        case(19): return "Industrial";
        case(20): return "Alternative";
        case(21): return "Ska";
        case(22): return "Death Metal";
        case(23): return "Pranks";
        case(24): return "Soundtrack";
        case(25): return "Euro-techno";
        case(26): return "Ambient";
        case(27): return "Trip-hop";
        case(28): return "Vocal";
        case(29): return "Jazz & Funk";
        case(30): return "Fusion";
        case(31): return "Trance";
        case(32): return "Classical";
        case(33): return "Instrumental";
        case(34): return "Acid";
        case(35): return "House";
        case(36): return "Game";
        case(37): return "Sound Clip";
        case(38): return "Gospel";
        case(39): return "Noise";
        case(40): return "Alternative Rock";
        case(41): return "Bass";
        case(42): return "Soul";
        case(43): return "Punk";
        case(44): return "Space";
        case(45): return "Meditative";
        case(46): return "Instrumental Pop";
        case(47): return "Instrumental Rock";
        case(48): return "Ethnic";
        case(49): return "Gothic";
        case(50): return "Darkwave";
        case(51): return "Techno-Industrial";
        case(52): return "Electronic";
        case(53): return "Pop-Folk";
        case(54): return "Eurodance";
        case(55): return "Dream";
        case(56): return "Southern Rock";
        case(57): return "Comedy";
        case(58): return "Cult";
        case(59): return "Gangsta Rap";
        case(60): return "Top 40";
        case(61): return "Christian Rap";
        case(62): return "Pop/Funk";
        case(63): return "Jungle";
        case(64): return "Native American";
        case(65): return "Caberet";
        case(66): return "New Wave";
        case(67): return "Psychadelic";
        case(68): return "Rave";
        case(69): return "Showtunes";
        case(70): return "Trailer";
        case(71): return "Lo-Fi";
        case(72): return "Tribal";
        case(73): return "Acid Punk";
        case(74): return "Acid Jazz";
        case(75): return "Polka";
        case(76): return "Retro";
        case(77): return "Musical";
        case(78): return "Rock & Roll";
        case(79): return "Hard Rock";
        /* Winamp defined genres
           www.winamp.com     */
        case(80): return "Folk";
        case(81): return "Folk-Rock";
        case(82): return "National Folk";
        case(83): return "Swing";
        case(84): return "Fast Fusion";
        case(85): return "Bebob";
        case(86): return "Latin";
        case(87): return "Revival";
        case(88): return "Celtic";
        case(89): return "Bluegrass";
        case(90): return "Avantgarde";
        case(91): return "Gothic Rock";
        case(92): return "Progressive Rock";
        case(93): return "Psychedelic Rock";
        case(94): return "Symphonic Rock";
        case(95): return "Slow Rock";
        case(96): return "Big Band";
        case(97): return "Chorus";
        case(98): return "Easy Listening";
        case(99): return "Acoustic";
        case(100): return "Humor";
        case(101): return "Speech";
        case(102): return "Chanson";
        case(103): return "Opera";
        case(104): return "Chamber Music";
        case(105): return "Sonata";
        case(106): return "Symphony";
        case(107): return "Booty Bass";
        case(108): return "Primus";
        case(109): return "Porn Groove";
        case(110): return "Satire";
        case(111): return "Slow Jam";
        case(112): return "Club";
        case(113): return "Tango";
        case(114): return "Samba";
        case(115): return "Folklore";
        case(116): return "Ballad";
        case(117): return "Power Ballad";
        case(118): return "Rythmic Soul";
        case(119): return "Freestyle";
        case(120): return "Duet";
        case(121): return "Punk Rock";
        case(122): return "Drum Solo";
        case(123): return "Acapella";
        case(124): return "Euro-House";
        case(125): return "Dance Hall";
    }
    return "Unknown genre";
}
#endif

int parse_args(int argc, char *argv[]) {
    FILE *testopen;
    int retval = 1;
    createconfig = 1;

    if(argc < 2)
        return 1;
        
    switch(argv[1][0]) {
        case('-'):
            if(strlen(argv[1]) != 2) {
                fprintf(stderr, "MSP does not recognize the switch \"%s\".\n\n", argv[1]);
                return 1;
            }
            break;
        default:
            fprintf(stderr, "MSP does not recognize the switch \"%s\".\n\n", argv[1]);
            return 1;
    }
    
    switch(argv[1][1] & 223) {
        case('I'):
            return 2;
        case('\x1f'): /* '?' & 223 */
            return 1;
        case('W'):
            retval = 3;
            break;
        case('M'):
            retval = 4;
            break;
        case('C'):
            retval = 5;
            break;
#ifndef ANOREXIC
        case('A'):
            retval = 6;
            break;
#endif
        case('T'):
            retval = 7;
            break;
        case('G'):
            retval = 8;
            break;
#ifndef ANOREXIC        /* This line contained #ifdef in v0.50, thus breaking MOD playing in the normal binary */
        case('D'):      /* Note to self: Test programs better before relase. -Sevy ;-) */
            retval = 9;
            break;
#endif
        case('P'):
            retval = 10;
            break;
        case('N'):
            retval = 11;
            break;
        default:
            fprintf(stderr, "MSP does not recognize the switch \"%s\".\n\n", argv[1]);
            return 1;
    }
    
    /* Check the second arg */
    if(argc < 3) {
        if(retval == 7)
            fprintf(stderr, "Please specify a tone number.\n\n");
        else if(retval == 10)
            fprintf(stderr, "Please specify a CD track number.\n\n");
        else
            fprintf(stderr, "Please specify a file name.\n\n");
        return 1;
    }

    switch(retval) {
        case(7):
            if(!atoi(argv[2])) {
                fprintf(stderr, "Invalid tone number \"%s\"\n\n", argv[2]);
                return 1;
            }
            return retval;
        case(10):
            if(!atoi(argv[2])) {
                fprintf(stderr, "Invalid CD track number \"%s\"\n\n", argv[2]);
                return 1;
            }
            return retval;
        case(11):
            if(argc != 4) {
                fprintf(stderr, "Please specify a track number.\n\n");
                return 1;
            }
            return retval;
        default:
            break;
    }

    testopen = fopen(argv[2], "rb");
    if(!testopen) {
        fprintf(stderr, "Error opening ");
        /* Prints an informative error message */
        perror(argv[2]);
        fputc('\n', stderr);
        return 1;
    }
    fclose(testopen);
    if(argc == 4) {
        switch(argv[3][0]) {
            case('-'):
                if(strlen(argv[3]) != 2) {
                    fprintf(stderr, "MSP does not recognize the switch \"%s\".\n\n", argv[3]);
                    return 1;
                }
                break;
            default:
                fprintf(stderr, "MSP does not recognize the second switch \"%s\".\n\n", argv[3]);
                return retval; /* Let's the file still be played if a bad switch is used for looping */
        }
        switch(argv[3][1] & 223) {
            case('L'):
                wantloop = 1;
                break;
            case('X'):
                createconfig = 0;
                break;
            default:
                fprintf(stderr, "MSP does not recognize the second switch \"%s\".\n\n", argv[3]);
                break;
        }
    }
    return retval;
}

/* All you need to know about how to use MSP ;-) */
void show_help(char *progname) {
    printf("Usage: %s [opt] [file, Hz (tones), or CD track #] [2nd opt or NSF track #]\n\n", strrchr(progname, '/') + 1);
    printf("Options:    -? = Help = This Message!\n");
#ifndef ANOREXIC
    printf("            -a = play .MP3 or .MP2 file\n");
#endif
    printf("            -c = play .CYM file (Callus YM2151 Dump Files)\n");
#ifndef ANOREXIC
    printf("            -d = play .MOD, .XM, .ULT, .M15, .S3M, .UNI, .MTM, or .STM file\n");
#endif
    printf("            -g = play .GYM file (Genecyst YM2612/SN76496 Dump Files)\n");
    printf("            -i = Information About MSP\n");
    printf("            -w = play .WAV or .VOC file (only mono VOC files!)\n");
    printf("            -m = play .MID file\n");
    printf("            -n = play .NSF file (NES Sound Format)\n");
    printf("            -p = play CD track x using MSCDEX\n");
    printf("            -t = generate tone of x Hz (with PC speaker)\n");
    printf("2nd Option: -l = loop sound file (continuous play)\n");
    printf("            -x = suppress the creation of MSP.CFG\n");
    return;
}

void show_info(void) {
    printf("Compiled %s with GCC %i.%i (djgpp) %s\n\n", __DATE__, __GNUC__, __GNUC_MINOR__, __TIME__);
    allegro_init();
    check_cpu();
    switch(os_type) {
        case(OSTYPE_WIN98):
            printf("Running on a %s processor with Windows 98...\n\n", cpu_vendor);
            break;
        case(OSTYPE_WIN95):
            printf("Running on a %s processor with Windows 95...\n\n", cpu_vendor);
            break;
        case(OSTYPE_WIN3):
            printf("Running on a %s processor with Windows 3.x...\n\n", cpu_vendor);
            break;
        case(OSTYPE_WINNT):
            printf("Running on a %s processor with Windows NT...\n\n", cpu_vendor);
            break;
        case(OSTYPE_OS2):
            printf("Running on a %s processor with OS/2...\n\n", cpu_vendor);
            break;
        case(OSTYPE_WARP):
            printf("Running on a %s processor with OS/2 (WARP)...\n\n", cpu_vendor);
            break;
        case(OSTYPE_OPENDOS):
            printf("Running on a %s processor with OpenDOS...\n\n", cpu_vendor);
            break;
        case(OSTYPE_DOSEMU):
            printf("Running on a %s processor with DOSEMU (Linux, etc.)...\n\n", cpu_vendor);
            break;
        default:
            printf("Running on a %s processor with MS-DOS or an unknown OS...\n\n", cpu_vendor);
            break;
    }
    printf("Credits, etc.:\n");
    printf("YM2151, YM2612, and SN76496 Code (C) The M.A.M.E. Team, Jarek Burczynski, etc.\n");
    printf("MP3 File support from LibAmp v0.3, an excellent MP3 library for DJGPP/Allegro.\n");
    printf("LibAmp, by Ove Kaaven, is a port of Amp 0.7.6 by Tomislav Uzelac.\n");
    printf("MOD, etc. support from MikAllegro, a port of MikMod to djgpp/Allegro.\n");
    printf("NSF support from Nosefart 1.7 by Matthew Conte.\n");
    printf("CD playing code originally by Steven H. Don; modified by Sevy and Marp.\n");
    printf("Thanx to Shawn Hargreaves for Allegro, an excellent game library for djgpp.\n");
    printf("EXE compression by UPX.\n");
    printf("GYM decompression by zlib.\n");
    printf("All other l33t C code by Sevy and Marp.\n");
    printf("Please read README.TXT\n");
    srand(time(0));
    printf("\nLucky lotto numbers: %i %i %i %i %i %i\n", rand()%100, rand()%100, rand()%100, rand()%100, rand()%100, rand()%100);
    return;
}

int play_wav_file(char *filename) {
    SAMPLE *the_sample;
    int pan = 128;
    int pitch = 1000;

    allegro_init();
    if(install_sound(DIGI_AUTODETECT, MIDI_NONE, NULL)) {
        fprintf(stderr, "Sound initialization error: %s\n", allegro_error);
        return 1;
    }
    the_sample = load_sample(filename);
    if(!the_sample) {
        fprintf(stderr, "There was an error while reading the WAV file '%s'.\n", filename);
        return 1;
    }
    printf("Digital Sound Driver Installed: %s\n", digi_driver->name);
    printf("MSP is now playing: %s (%li bytes)\n", filename, file_size(filename));
    printf("Press any key to stop playing or to quit.\n");
    play_sample(the_sample, 255, pan, pitch, wantloop);
    getch(); /* Wait until the user presses a key */
    destroy_sample(the_sample); /* Stop playing */
    while(kbhit())
        getch(); /* Clear the keyboard buffer */
    return 0;
}

int play_midi_file(char *filename) {
    MIDI *midi_file;

    allegro_init();
    install_timer();
    if(install_sound(DIGI_AUTODETECT, MIDI_AUTODETECT, NULL)) {
        fprintf(stderr, "Sound initialization error: %s\n", allegro_error);
        return 1;
    }
    midi_file = load_midi(filename);
    if(!midi_file) {
        fprintf(stderr, "There was an error while reading the MIDI file '%s'.\n", filename);
        return 1;
    }
    printf("MIDI Sound Driver Installed: %s\n", midi_driver->name);
    printf("MSP is now playing: %s (%li bytes)\n", filename, file_size(filename));
    printf("Press any key to stop playing or to quit.\n");
    play_midi(midi_file, wantloop); /* Play the file */
    getch(); /* Wait until the user presses a key */
    destroy_midi(midi_file); /* Stops playing the file */
    while(kbhit())
        getch(); /* Clear the keyboard buffer */
    return 0;
}

int play_cym_file(char *filename) {
    FILE *cym_file;
    AUDIOSTREAM *stream;
    unsigned long loop, loop2, rate, rate15, rate30, YM2151clock, wantstereo,
        cym_filelength = 0, data_position = 0, timetoupdate = 0;
    unsigned char *cym_data;
    unsigned short *ym_buffer[2], *stream_buf;

    allegro_init();
    rate = get_config_int(NULL, "samprate", 48000);
    switch(rate) {
        case(48000):
            break;
        case(44100):
            break;
        case(24000):
            break;
        case(22050):
            break;
        default:
            fprintf(stderr, "Error: invalid sample rate %lu.\n", rate);
            return 0;
    }
    rate15 = rate / 15; /* number of bytes per update */
    rate30 = rate / 30; /* number of samples per update */
    wantstereo = get_config_int(NULL, "stereo", 1);
    if(install_sound(DIGI_AUTODETECT, MIDI_NONE, NULL)) {
        fprintf(stderr, "Sound initialization error: %s\n", allegro_error);
        return 1;
    }
    stream = play_audio_stream(rate30, 16, wantstereo, rate, 255, 128);
    if(!stream) {
        fprintf(stderr, "There was a problem with the audio stream.\n");
        return 1;
    }
    printf("Digital Sound Driver Installed: %s\n", digi_driver->name);
    cym_filelength = file_size(filename);
    cym_file = fopen(filename, "rb"); /* Open the file */
    if(!cym_file) { /* Makes sure the file exists, duh! */
        fprintf(stderr, "\nError opening ");
        perror(filename);
        return 1;
    }
    cym_data = malloc(cym_filelength); /* Allocates memory for the size of the CYM file */
    fread(cym_data, 1, cym_filelength, cym_file); /* Reads the file */
    fclose(cym_file); /* Closes the file when done. */
    ym_buffer[0] = (unsigned short *)malloc(rate15);
    ym_buffer[1] = (unsigned short *)malloc(rate15);
    YM2151clock = get_config_int(NULL, "YM2151clock", 3579545);
    YM2151Init(1, YM2151clock, rate);
    printf("YM2151 Loaded...\n");
    printf("MSP is now playing: %s (%lu bytes)\n", filename, cym_filelength);
    printf("Press any key to stop playing or to quit.\n");
    while(!kbhit() && data_position < cym_filelength) { /* Does the stuff to play the file until you press a key. (with looping) */
        if(cym_data[data_position] == '\0') {
            if(timetoupdate) {
                /* 1/30 sec has passed. Time to update everything */
                timetoupdate = 0;
                YM2151UpdateOne(0, (void **)ym_buffer, rate30);
                data_position++;
                stream_buf = get_audio_stream_buffer(stream);
                while(!stream_buf) {
                    __dpmi_yield();
                    stream_buf = get_audio_stream_buffer(stream);
                }
                if(wantstereo) {
                    loop2 = 0;
                    for(loop = 0; loop < rate30; loop++) {
                        stream_buf[loop2] = ym_buffer[0][loop] ^ 0x8000;
                        loop2++;
                        stream_buf[loop2] = ym_buffer[1][loop] ^ 0x8000;
                        loop2++;
                    }
                }
                else {
                    for(loop = 0; loop < rate30; loop++)
                        stream_buf[loop] = (((int)ym_buffer[0][loop] + ym_buffer[1][loop]) / 2) ^ 0x8000;
                }
                free_audio_stream_buffer(stream);
            }
            else
                timetoupdate = 1;
        }
        else {
            YM2151WriteReg(0, cym_data[data_position], cym_data[data_position + 1]);
            data_position += 2;
        }
        if(wantloop && data_position >= cym_filelength) {
            data_position = 0;
            YM2151ResetChip(0);
        }
    }
    stop_audio_stream(stream);
    YM2151Shutdown();
    free(cym_data);
    free(ym_buffer[0]);
    free(ym_buffer[1]);
    while(kbhit())
        getch(); /* Clear the keyboard buffer */
    return 0;
}

#ifndef ANOREXIC
int play_mp3_file(char *filename) {
    FILE *mp3_file;
    MP3TAG mp3tag;

    mp3_file = fopen(filename, "rb");
    if(!mp3_file) {
        fprintf(stderr, "Error opening ");
        perror(filename);
        return -1;
    }
    fseek(mp3_file, -128, SEEK_END);
    fread(&mp3tag, 128, 1, mp3_file);
    fclose(mp3_file);
    allegro_init();
    if(install_sound(DIGI_AUTODETECT, MIDI_NONE, NULL)) {
        printf("Sound initialization error: %s\n", allegro_error);
        return 1;
    }
    printf("Digital Sound Driver Installed: %s\n", digi_driver->name);
    install_amp();
    printf("MP3 Decoding Engine Successfully Loaded...\n");
    load_amp(filename, wantloop);
    if(!amp_bitrate || !amp_samprat) { /* I made this up so that it would generically figure out if the file is a valid MP3 */
        printf("There was an error while reading the MP3 file '%s'.\n", filename);
        return 1;
    }
    printf("MSP is now playing: %s (%li bytes)\n\n", filename, file_size(filename));
    printf("File Info:\t%dkbps/%dkHz\n", amp_bitrate, amp_samprat); /* Displays file info */
    printf("Stereo:\t\t%s\n\n", amp_stereo ? "YES" : "NO");
    if(!strncmp(mp3tag.ID, "TAG", 3)) { /* Check for ID3 tag. */
        printf("Song title:\t%.30s\n", mp3tag.title);
        printf("Artist:\t\t%.30s\n", mp3tag.artist);
        printf("Album:\t\t%.30s\n", mp3tag.album);
        printf("Year:\t\t%.4s\n", mp3tag.year);
        printf("Comment:\t%.30s\n", mp3tag.comment);
        if(!mp3tag.comment[28] && mp3tag.comment[29])
            printf("Album track:\t%i\n", mp3tag.comment[29]);
        printf("Genre:\t\t%s\n\n", get_genre_string(mp3tag.genre));
    }
    printf("Press any key to stop playing or to quit.\n");
    while((run_amp() >= 0) && !kbhit())
        __dpmi_yield();
    unload_amp();
    while(kbhit())
        getch();  /* Clear the keyboard buffer */
    return 0;
}
#endif

int play_tone(const char *frequency) {
    int freq = atoi(frequency);
    if(freq <= 0) {
        fprintf(stderr, "MSP cannot generate the specified tone!\n");
        return 1;
    }
    printf("MSP is playing a %i Hz tone.\n", freq);
    printf("Press any key to stop playing or to quit.\n");
    sound(freq);
    getch();
    nosound();
    while(kbhit())
        getch(); /* Clear the keyboard buffer */
    return 0;
}

int play_gym_file(char *filename) {
    unsigned long data_position, loop, loop2, gym_filelength, rate, YM2612clock,
        SNClock, wantstereo, dac_max, rate15, rate30, timetoupdate;
    unsigned short *stream_buf, *ym_buffer[2], *sn_buffer, *stream_ptr[2];
    unsigned char *gym_data, *dac_data, *tmp_data;
    FILE *gym_file;
    AUDIOSTREAM *stream;
    GYMTAG gymtag;

    allegro_init();
    rate = get_config_int(NULL, "samprate", 48000);
    switch(rate) {
        case(48000):
            break;
        case(44100):
            break;
        case(24000):
            break;
        case(22050):
            break;
        default:
            fprintf(stderr, "Error: invalid sample rate %li.\n", rate);
            return 0;
    }
    rate15 = rate / 15;
    rate30 = rate / 30;
    if(install_sound(DIGI_AUTODETECT, MIDI_NONE, NULL)) {
        fprintf(stderr, "Sound initialization error: %s\n", allegro_error);
        return 1;
    }
    wantstereo = get_config_int(NULL, "stereo", 1);
    stream = play_audio_stream(rate30, 16, wantstereo, rate, 255, 128);
    if(!stream) {
        fprintf(stderr, "There was a problem with the audio stream.\n");
        return 1;
    }
    printf("Digital Sound Driver Installed: %s\n", digi_driver->name);
    gym_filelength = file_size(filename);
    gym_file = fopen(filename, "rb");
    if(!gym_file) {
        fprintf(stderr, "Error opening ");
        perror(filename);
        return 1;
    }
    fread(&gymtag, 1, 428, gym_file);
    if(gymtag.id == 0x584d5947) {
        gym_filelength -= 428;
        tmp_data = (unsigned char *)malloc(gym_filelength);
        if(gymtag.compressed) {
#ifdef ANOREXIC
            fclose(gym_file);
            free(tmp_data);
            fprintf(stderr, "Error: compressed gym files not supported (anorexic build).\n");
            return 0;
#else
            gym_data = (unsigned char *)malloc(gymtag.compressed);
            fread(tmp_data, 1, gym_filelength, gym_file);
            uncompress(gym_data, &gymtag.compressed, tmp_data, gym_filelength);
            free(tmp_data);
            gym_filelength = gymtag.compressed;
#endif
        }
        else {
            gym_data = tmp_data;
            fread(gym_data, 1, gym_filelength, gym_file);
        }
    }
    else {
        fseek(gym_file, 0, SEEK_SET);
        gym_data = (unsigned char *)malloc(gym_filelength);
        fread(gym_data, 1, gym_filelength, gym_file);
    }
    fclose(gym_file);
    ym_buffer[0] = (unsigned short *)malloc(rate15);
    ym_buffer[1] = (unsigned short *)malloc(rate15);
    sn_buffer = (unsigned short *)malloc(rate15);
    dac_data = (unsigned char *)malloc(rate30);
    YM2612clock = get_config_int(NULL,"YM2612clock", 7670442);
    YM2612Init(1, YM2612clock, rate, NULL, NULL);
    SNClock = get_config_int(NULL, "SN76496clock", 3579580);
    SN76496_init(0, SNClock, rate);
    printf("YM2612 Loaded...\n");
    printf("SN76496 Loaded...\n");
    printf("MSP is now playing: %s (%lu bytes)\n", filename, gym_filelength);
    if(gymtag.id == 0x584d5947) {
        printf("\n     Song Name: %.32s\n", gymtag.songtitle);
        printf("     Game Name: %.32s\n", gymtag.gametitle);
        printf("Game Publisher: %.32s\n", gymtag.gamepub);
        printf("   Dumped With: %.32s\n", gymtag.dumpemu);
        printf("     Dumped By: %.32s\n", gymtag.dumper);
        printf("      Comments: %.256s\n\n", gymtag.comments);
    }
    printf("Press any key to stop playing or to quit.\n");
    data_position = 0;
    dac_max = 0;
    timetoupdate = 0;
    while(!kbhit() && data_position < gym_filelength) {
        switch(gym_data[data_position]) {
            case(0x00):
                data_position++;
                if(timetoupdate) {
                    /* 1/30 sec passed. update everything. */
                    timetoupdate = 0;
                    if(dac_max) {
                        double stream_cnt = 0;
                        double update_cycle = rate30 / dac_max; /* dac port write cycle */
                        for(loop = 0; loop < dac_max; loop++) {
                            int old_cnt = stream_cnt;
                            int step;                       /* update DAC port */
                            YM2612Write(0, 0, 0x2a);
                            YM2612Write(0, 1, dac_data[loop]);
                            if(loop == dac_max-1)
                                step = rate30 - old_cnt;
                            else {
                                stream_cnt += update_cycle;
                                step = (int)stream_cnt - old_cnt;
                            }
                            /* update stream */
                            stream_ptr[0] = &ym_buffer[0][old_cnt];
                            stream_ptr[1] = &ym_buffer[1][old_cnt];
                            YM2612UpdateOne(0, (void **)stream_ptr, step);
                        }
                        dac_max = 0;
                    }
                    else
                        YM2612UpdateOne(0, (void **)ym_buffer, rate30);
                    SN76496Update_16(0, sn_buffer, rate30);
                    while(!(stream_buf = get_audio_stream_buffer(stream)))
                        __dpmi_yield();
                    if(wantstereo) {
                        loop2 = 0;
                        for(loop = 0; loop < rate30; loop++) {
                            stream_buf[loop2] = sn_buffer[loop] + (ym_buffer[0][loop] ^ 0x8000);
                            loop2++;
                            stream_buf[loop2] = sn_buffer[loop] + (ym_buffer[1][loop] ^ 0x8000);
                            loop2++;
                        }
                    }
                    else {
                        for(loop = 0; loop < rate30; loop++)
                            stream_buf[loop] = sn_buffer[loop] + ((ym_buffer[0][loop] + ym_buffer[1][loop]) ^ 0x8000);
                    }
                    free_audio_stream_buffer(stream);
                }
                else
                    timetoupdate = 1;
                break;
            case(0x01):
                data_position++;
                if(gym_data[data_position] == 0x2a) {
                    data_position++;
                    if(dac_max < rate30)
                        dac_data[dac_max++] = gym_data[data_position];
                    data_position++;
                }
                else {
                    YM2612Write(0, 0, gym_data[data_position++]);
                    YM2612Write(0, 1, gym_data[data_position++]);
                }
                break;
            case(0x02):
                data_position++;
                YM2612Write(0, 2, gym_data[data_position]);
                data_position++;
                YM2612Write(0, 3, gym_data[data_position]);
                data_position++;
                break;
            case(0x03):
                data_position++;
                SN76496Write(0, gym_data[data_position]);
                data_position++;
                break;
            default:
                /* Prevents invalid or corrupt .gym files from hanging msp */
                data_position++;
                break;
        }
        if(wantloop && data_position >= gym_filelength) {
            data_position = 0;
            YM2612ResetChip(0);
        }
    }
    YM2612Shutdown();
    free(gym_data);
    free(ym_buffer[0]);
    free(ym_buffer[1]);
    free(sn_buffer);
    while(kbhit())
        getch(); /* Clear the keyboard buffer */
    return 0;
}

#ifndef ANOREXIC
/* Stuff for mod's, etc.
   This stuff uses the MikAllegro library for djgpp (MikMod)
   Some ideas were mooched from test programs, etc. ;-) */

void Player(void) { /* Player routine called @ a regular interval (beats-per-minute) */
    MP_HandleTick();
    MD_SetBPM(mp_bpm); /* Sets beats-per-minute */
}

int play_module_file(char *filename) {
    UNIMOD *mod_file;
    int wantstereo;

    wantstereo = get_config_int(NULL, "stereo", 1);
    md_mixfreq=get_config_int(NULL, "samprate", 48000);
    md_dmabufsize=(md_mixfreq / 8);
    if(wantstereo)
        md_mode=DMODE_STEREO|DMODE_16BITS|DMODE_INTERP;
    else
        md_mode=DMODE_16BITS|DMODE_INTERP;
    md_device=0;
    allegro_init();
    install_timer();
    /* Enables support for a bunch of file types */
    ML_RegisterLoader(&load_m15); /* M15 must be registered first */
    ML_RegisterLoader(&load_mod); /* 8 New File Types supported by MSP :-) */
    ML_RegisterLoader(&load_s3m);
    ML_RegisterLoader(&load_uni);
    ML_RegisterLoader(&load_mtm);
    ML_RegisterLoader(&load_stm);
    ML_RegisterLoader(&load_ult);
    ML_RegisterLoader(&load_xm);
    /* Enable sound card drivers */
    MD_RegisterDriver(&drv_nos); /* No Sound Driver */
    MD_RegisterDriver(&drv_ss); /* Ensoniq Soundscape Driver */
    MD_RegisterDriver(&drv_sb); /* SoundBlaster Driver */
    MD_RegisterDriver(&drv_gus); /* Gravis UltraSound Driver */
    MD_RegisterPlayer(Player); /* Points to the Player routine to update music */
    if(!MD_Init()) { /* Initializes sound card (obviously) */
        fprintf(stderr, "Sound initialization error: %s\n", myerr);
        return 1;
    }
    mod_file = ML_LoadFN(filename);
    if(!mod_file) {
            fprintf(stderr, "Error opening %s: %s\n", filename, myerr);
            return 1;
    }
    printf("Digital Sound Driver Installed: %s\n", md_driver->Name);
    printf("MSP is now playing: %s (%li bytes)\n", filename, file_size(filename));
    printf("Song Name: %s\n", mod_file->songname);
    printf("Module Type: %s\n", mod_file->modtype);
    printf("Comments: %s\n", mod_file->comment);
    printf("Press any key to stop playing or to quit.\n");
    mp_loop = wantloop;
    MP_Init(mod_file);
    md_numchn=mod_file->numchn;
    MD_PlayStart();
    install_int(MD_Update, 10);
    while(!kbhit() && !MP_Ready())
        __dpmi_yield();
    MD_PlayStop();
    ML_Free(mod_file);
    MD_Exit();
    while(kbhit())
        getch(); /* Clear the keyboard buffer */
    return 0;
}
#endif

/* NSF (NES Sound Format) code added by Matthew Conte (Nofrendo rules!!!) ;-) */

/* one time init */
static void nsf_setup(nsf_t *nsf) {
    log_init(); /* i love you, hercules! */
    log_printf("Starting Nosefart engine v1.7\n");
    nes_init(nsf);
    log_printf("\n- Host initialization complete\n\n");
    return;
}

/* one time shutdown */
static void nsf_shutdown(void) {
    nes_shutdown();
    log_printf("- Shutting down host machine\n");
    log_shutdown();
    return;
}

/* set up a track to play */
static void nsf_setupsong(nsf_t *nsf) {
    printf("Playing NSF track %d of %d\t\r", nsf->current_song, nsf->num_songs);
    nes_reset(HARD_RESET);
    return;
}

/* display info about an NSF file */
static void nsf_displayinfo(nsf_t *nsf) {
    printf("\nKeys: (esc to quit)\n");
    printf(" - left/right arrows to change track\n");
    printf(" - return restarts track\n");
    printf(" - 1-6 set toggle channels on/off\n");
    printf(" - A/S/D set DSP level [none / low / high]\n\n");

    printf("NSF Info:\n");
    printf("\tSong name: \t%s\n", nsf->song_name);
    printf("\tAuthor:    \t%s\n", nsf->author_name);
    printf("\tCopyright: \t%s\n", nsf->copyright);
    printf("\tNum. Songs:\t%d\n\n", nsf->num_songs);
    return;
}

/* set a channel to on/off */
void set_channel(int channel) {
    static int chan_enabled[6] = { 1, 1, 1, 1, 1, 1 };

    chan_enabled[channel] ^= 1;
    apu_setchan(channel, chan_enabled[channel]);
    return;
}

/* handle keypresses */
static int nsf_handlekey(nsf_t *nsf) {
    char ch;

    ch = getch();
    /* special naughty case for extkeys - bleh! */
    if(0 == ch)
        ch = getch();
    switch (ch) {
        case(27): /* esc */
            printf("\n");
            return 1;
        case(77): /* right cursor arrow */
            if(nsf->current_song == nsf->num_songs)
                break;
            nsf->current_song++;
            nsf_setupsong(nsf);
            break;
        case(75): /* left cursor arrow */
            if(1 == nsf->current_song)
                break;
            nsf->current_song--;
            nsf_setupsong(nsf);
            break;
        case(13): /* return */
            /* nsf_setupsong(nsf); */
            nes_reset(HARD_RESET);
            break;
        case(49): /* 1 */
        case(50): /* 2 */
        case(51): /* 3 */
        case(52): /* 4 */
        case(53): /* 5 */
        case(54): /* 6 */
            set_channel(ch - 49);
            break;
        case(97):
        case(65): /* A/a */
            apu_setdsp(0);
            break;
        case(115):
        case(83): /* S/s */
            apu_setdsp(1);
            break;
        case(100):
        case(68):
            apu_setdsp(2);
            break;
        default:
            break;
    }
    return 0;
}

/* play that thar NSF file */
static int play_nsf_file(char *filename, int track) {
    int rate, bufsize, done = 0;
    unsigned short *stream_buf;
    AUDIOSTREAM *stream;
    nsf_t *nsf;

    /* init allegro monster */
    allegro_init();
    if(install_sound(DIGI_AUTODETECT, MIDI_NONE, NULL)) {
        fprintf(stderr, "Sound initialization error: %s\n", allegro_error);
        return 1;
    }

    /* get desired sampling rate from config file */
    rate = get_config_int(NULL, "samprate", 48000);

    /* ha ha, 16-bit nonsense */
    apu_init(rate, 16, 1);

    /* load up an NSF file */
    nsf = nsf_load(filename);
    if (!nsf) {
        fprintf(stderr, "Error opening ");
        perror(filename);
        return 1;
    }

    /* calculate desired buffer size from NSF playback rate (in hertz) */
    bufsize = (rate / nsf->playback_rate);

    /* set up the nsf */
    nsf_setup(nsf);

    /* determine which track to play */
    if (track > nsf->num_songs || track < 1) {
        nsf->current_song = nsf->start_song;
        fprintf(stderr, "track %d out of range, playing track %d\n", track, nsf->current_song);
    }
    else
       nsf->current_song = track;

    /* always mono! */
    stream = play_audio_stream(bufsize, 16, 0, rate, 255, 128);
    if (!stream) {
        fprintf(stderr, "There was a problem with the audio stream.\n");
        return 1;
    }
    printf("Digital Sound Driver Installed: %s\n", digi_driver->name);
    printf("MSP is now playing: %s (%li bytes)\n", filename, file_size(filename));

    /* display file info and reset state of the NES */
    nsf_displayinfo(nsf);
    nsf_setupsong(nsf);

    /* main apu processing loop */
    while(!done) {
        /* do a frame's worth of emulation (until kill switch is hit) */
        nes_frame();

        /* oh god mr. hargreaves!  what filth is this?!?! */
        stream_buf = get_audio_stream_buffer(stream);
        while(!stream_buf) {
            __dpmi_yield();
            stream_buf = get_audio_stream_buffer(stream);
        }

        /* render some samples */
        apu_process16(stream_buf, bufsize);

        free_audio_stream_buffer(stream);

        if(kbhit())
            done = nsf_handlekey(nsf);
    }

    /* shutdown everything we've laid our filthy hands upon */
    stop_audio_stream(stream);
    nsf_free(nsf);
    apu_shutdown();
    nsf_shutdown();

    while(kbhit())
        getch(); /* Clear the keyboard buffer */

    return 0;
}


/* Function to play a CD track */
int play_cd_track(const char *tracknumber) {
    int numberoftracks;

    /* Looks for MSCDEX and at least one CDROM player */
    if(!Check_MSCDEX()) {
        printf("MSCDEX not installed or no CD-ROM drive available.\n");
        return 0;
    }
    else
        /* A function in cdplayer.c will show the MSCDEX Version number before this */
        printf("CD-ROM installed as drive %c:\n", Drive_Letter + 65);

    GetAudioCDInfo();
    if(!CheckForErrors()) {
        numberoftracks = (Last_Track - First_Track + 1);
        printf("There are %i tracks on this CD\n" , numberoftracks);
        if((atoi(tracknumber)) < First_Track || (atoi(tracknumber)) > Last_Track) {
            printf("The track you specified, %s, is not on this CD.\n", tracknumber);
            return 0;
        }
        printf("Total playing time: %ldmin %ld.%ldsec\n", LeadOut_Track >> 16, (LeadOut_Track & 0xFF00) >> 8, (100*LeadOut_Track & 0xFF) / 75);
        TrackToPlay = atoi(tracknumber);
        printf("MSP is now playing track number: %s\n", tracknumber);
        printf("Press any key to stop playing or to quit.\n");
        PlayTrack(TrackToPlay);
        /* CheckForErrors(); */
        getch();
        StopPlaying();
        while(kbhit())
            getch(); /* Clear the keyboard buffer */
    }
    return 0;
}

/* Function to create a configuration file for MSP */
int create_config_file(void) {
    FILE *config_file;

    if(!createconfig)
        return 0;

    config_file = fopen("msp.cfg", "w");
    if(!config_file) {
        /* Prints an informative error message */
        perror("Error creating msp.cfg");
        return 1;
    }
    fputs("# MSP v0.51 Configuration File\n", config_file);
    fputs("# Edit the values in this file as you would like :-)\n\n", config_file);
    fputs("loop = 0\n", config_file);
    fputs("# 0 == No Loop, Non-Zero == Loop (default: 0)\n", config_file);
    fputs("# Note:  Over-rides command line switch '-l'\n\n", config_file);
    fputs("samprate = 48000\n", config_file);
    fputs("# Sample Rate for GYM, CYM, and MOD file playing (default: 48000)\n", config_file);
    fputs("# Valid rates are 48000, 44100, 24000 and 22050\n\n", config_file);
    fputs("stereo = 1\n", config_file);
    fputs("# Got stereo? 0 == Mono (not stereo), Non-Zero == Stereo (default: Stereo)\n\n", config_file);
    fputs("YM2612clock = 7670442\n", config_file);
    fputs("# The clock rate of the YM2612 Processor (default: 7670442)\n", config_file);
    fputs("# Note: Only mess with this if you really know what you're doing ;-)\n\n", config_file);
    fputs("SN76496clock = 3579580\n", config_file);
    fputs("# The clock rate of the SN76496 Processor (default: 3579580)\n", config_file);
    fputs("# Note: See Above\n\n", config_file);
    fputs("YM2151clock = 3579545\n", config_file);
    fputs("# The clock rate of the YM2151 Processor (default: 3579545)\n", config_file);
    fputs("# Note: See Above\n\n", config_file);
    return 0;
}

/* Obviously, this starts everything and the gets the command line args. */
int main(int argc, char *argv[]) {
    printf("\nMSP (Miscellaneous Sound Player) v0.51 by Sevy & Marp\n\n");

    if(exists("msp.cfg")) {
        override_config_file("msp.cfg");
        wantloop = get_config_int(NULL, "loop", 0);
    }

    switch(parse_args(argc, argv)) {
        case(2):
            show_info();
            break;
        case(3):
            if(!exists("msp.cfg"))
                create_config_file();
                return play_wav_file(argv[2]);
        case(4):
            if(!exists("msp.cfg"))
                create_config_file();
            return play_midi_file(argv[2]);
        case(5):
            if(!exists("msp.cfg"))
                create_config_file();
            return play_cym_file(argv[2]);
#ifndef ANOREXIC
        case(6):
            if(!exists("msp.cfg"))
                create_config_file();
            return play_mp3_file(argv[2]);
#endif
        case(7):
            if(!exists("msp.cfg"))
                create_config_file();
            return play_tone(argv[2]);
        case(8):
            if(!exists("msp.cfg"))
                create_config_file();
            return play_gym_file(argv[2]);
#ifndef ANOREXIC
        case(9):
            if(!exists("msp.cfg"))
                create_config_file();
            return play_module_file(argv[2]);
#endif
        case(10):
            if(!exists("msp.cfg"))
                create_config_file();
            return play_cd_track(argv[2]);
        case(11):
            if(!exists("msp.cfg"))
                create_config_file();
            return play_nsf_file(argv[2], atoi(argv[3]));
        default:
            show_help(argv[0]);
            break;
    } /* switch(parse_args) */
    return 0;
}
