/*
** Nofrendo - NES emulator
** (c) 1999 Matthew Conte
**
** License agreement:
**
** This source code may be distributed in unmodified form to any medium,
** and may be used in any *freeware* project.
**
** It may not be sold, in full or in part, as part of any commercial or
** shareware sofware package without the express written permission of 
** the author, Matthew Conte (matt@conte.com).
**
** Modified versions of this code may not be publicly distributed
** without the submission to, and approval of the author (matt@conte.com).
** This includes distribution in source, object, or binary form.
**
** The author claims no responsibility for damage cause by this code; it
** is distributed "as-is"; in other words, use at your own risk.
**
** Any freeware application using this code must credit the author,
** Matthew Conte, in the application itself or the documentation.
**
** Modifications should be reported to matt@conte.com, to ensure the
** highest degree of quality and accuracy is always freely available.
**
**
** nes6502.h
**
** NES custom 6502 CPU definitions / prototypes
*/

/* NOTE: 16-bit addresses avoided like the plague: use 32-bit values
**       wherever humanly possible
*/

#ifndef _NES6502_H_
#define _NES6502_H_

/* Define this for little-endian host architecture (x86) */
#define  NES6502_LITTLE_ENDIAN

/* Define this to enable decimal mode in ADC / SBC (not needed in NES) */
/*
#define  NES6502_DECIMAL
*/

/* P (flag) register bitmasks */
#define  N_FLAG         0x80
#define  V_FLAG         0x40
#define  R_FLAG         0x20  /* Reserved, always 1 */
#define  B_FLAG         0x10
#define  D_FLAG         0x08
#define  I_FLAG         0x04
#define  Z_FLAG         0x02
#define  C_FLAG         0x01

/* Vector addresses */
#define  NMI_VECTOR     0xFFFA
#define  RESET_VECTOR   0xFFFC
#define  IRQ_VECTOR     0xFFFE

#define  INTERRUPT_CYCLES  7

/* Stack is located on 6502 page 1 */
#define  STACK_PAGE     0x0100

typedef struct
{
   uint32 min_range;
   uint32 max_range;
   uint8 (*read_func)(uint32 address);
} nes6502_memread;

typedef struct
{
   uint32 min_range;
   uint32 max_range;
   void (*write_func)(uint32 address, uint8 value);
} nes6502_memwrite;

typedef struct
{
   uint8 *mem_page[8];   /* 8kB memory page pointers */
   nes6502_memread *read_handler;
   nes6502_memwrite *write_handler;
   uint32 pc_reg;
   uint8 a_reg, p_reg;
   uint8 x_reg, y_reg;
   uint8 s_reg;
   boolean irq_pending, nmi_pending;
#ifdef NES6502_DEBUG
   uint32 trap_addr;
   boolean trace;
#endif
} nes6502_context;

/* Functions which govern the 6502's execution */
extern void  nes6502_init(void);
extern void  nes6502_reset(void);
extern int   nes6502_execute(int total_cycles);
extern void  nes6502_nmi(void);
extern void  nes6502_irq(void);
extern uint8 nes6502_dma(uint32 address);
extern int   nes6502_getcycles(boolean reset_flag);

/* Context get/set */
extern void  nes6502_setcontext(nes6502_context *cpu);
extern void  nes6502_getcontext(nes6502_context *cpu);

#endif /* _NES6502_H_ */
