/*
** Nofrendo - NES emulator
** (c) 1999 Matthew Conte
**
** License agreement:
**
** This source code may be distributed in unmodified form to any medium,
** and may be used in any *freeware* project.
**
** It may not be sold, in full or in part, as part of any commercial or
** shareware sofware package without the express written permission of 
** the author, Matthew Conte (matt@conte.com).
**
** Modified versions of this code may not be publicly distributed
** without the submission to, and approval of the author (matt@conte.com).
** This includes distribution in source, object, or binary form.
**
** The author claims no responsibility for damage cause by this code; it
** is distributed "as-is"; in other words, use at your own risk.
**
** Any freeware application using this code must credit the author,
** Matthew Conte, in the application itself or the documentation.
**
** Modifications should be reported to matt@conte.com, to ensure the
** highest degree of quality and accuracy is always freely available.
**
**
** nes_apu.h
**
** NES APU emulation header file
*/

#ifndef _NES_APU_H_
#define _NES_APU_H_

#define  APU_WRA0    0x4000
#define  APU_WRA1    0x4001
#define  APU_WRA2    0x4002
#define  APU_WRA3    0x4003
#define  APU_WRB0    0x4004
#define  APU_WRB1    0x4005
#define  APU_WRB2    0x4006
#define  APU_WRB3    0x4007
#define  APU_WRC0    0x4008
#define  APU_WRC2    0x400A
#define  APU_WRC3    0x400B
#define  APU_WRD0    0x400C
#define  APU_WRD2    0x400E
#define  APU_WRD3    0x400F
#define  APU_WRE0    0x4010
#define  APU_WRE1    0x4011
#define  APU_WRE2    0x4012
#define  APU_WRE3    0x4013

#define  APU_OAMDMA  0x4014
#define  APU_SMASK   0x4015
#define  APU_JOY0    0x4016
#define  APU_JOY1    0x4017

#define  APU_INVALID 0xFF

#define  NOISE_LONG     0x4000
#define  NOISE_SHORT    93

typedef struct square_s
{
   uint8 regs[4];
   int vbl_length;
   int freq;
   float phaseacc;
   float output_vol;
   float env_phase;
   float sweep_phase;
   uint8 adder;
   uint8 env_vol;
   boolean enabled;
} square_t;

typedef struct triangle_s
{
   uint8 regs[4]; /* regs[1] unused */
   int linear_length;
   int vbl_length;
   int write_latency;
   float phaseacc;
   float output_vol;
   uint8 adder;
   boolean counter_started;
   boolean enabled;
} triangle_t;

typedef struct noise_s
{
   uint8 regs[4]; /* regs[1] unused */
   int cur_pos;
   int vbl_length;
   float phaseacc;
   float output_vol;
   float env_phase;
   uint8 env_vol;
   boolean enabled;
} noise_t;

typedef struct dpcm_s
{
   uint8 regs[4];
   uint32 address;
   uint32 length;
   int bits_left;
   float phaseacc;
   float output_vol;
   uint8 cur_byte;
   boolean enabled;
   boolean irq_occurred;
} dpcm_t;

/* apu ring buffer */
typedef struct apuqueue_s
{
   int timestamp;
   uint32 address;
   uint8 value;
} apuqueue_t;

#define APUQUEUE_SIZE  0x2000

enum
{
   APU_DSP_NONE,
   APU_DSP_SIMPLE,
   APU_DSP_SCALED
};

typedef int8 (*apusound_t)(void);

extern float apu_getphase(void);

/* Function prototypes */
extern void apu_init(int sample_frequency, int bps, int nch);
extern void apu_shutdown(void);
extern void apu_setext(apusound_t ext_sound);
extern void apu_setdsp(int dsp_type);
extern void apu_process8(uint8 *buffer, int size);
extern void apu_process16(uint16 *buffer, int size);
extern void apu_reset(void);
extern void apu_setchan(int chan, boolean enabled);

extern uint8 apu_read(uint32 address);
extern void apu_write(uint32 address, uint8 value);

extern void apu_getpcmdata(uint8 **data, int *length);

#endif /* _NES_APU_H_ */
