/*
** Nofrendo - NES emulator
** (c) 1999 Matthew Conte
**
** License agreement:
**
** This source code may be distributed in unmodified form to any medium,
** and may be used in any *freeware* project.
**
** It may not be sold, in full or in part, as part of any commercial or
** shareware sofware package without the express written permission of 
** the author, Matthew Conte (matt@conte.com).
**
** Modified versions of this code may not be publicly distributed
** without the submission to, and approval of the author (matt@conte.com).
** This includes distribution in source, object, or binary form.
**
** The author claims no responsibility for damage cause by this code; it
** is distributed "as-is"; in other words, use at your own risk.
**
** Any freeware application using this code must credit the author,
** Matthew Conte, in the application itself or the documentation.
**
** Modifications should be reported to matt@conte.com, to ensure the
** highest degree of quality and accuracy is always freely available.
**
**
** nsf.c
**
** NSF loading/saving related functions
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "types.h"
#include "nsf.h"
#include "log.h"

/* TODO: cleaner, damnit */
#include "nes_apu.h"
#include "vrcvisnd.h"

/* TODO: X register is set based on PAL/NTSC??? */
static void nsf_setup(nsf_t *nsf)
{
   uint8 bsregs[8];
   int div_speed;

   nsf->current_song = nsf->start_song;

   if (nsf->pal_ntsc_bits & 1)
      div_speed = nsf->pal_speed;
   else
      div_speed = nsf->ntsc_speed;

   if (div_speed)
      nsf->playback_rate = 1000000 / div_speed;
   else
      nsf->playback_rate = 60; /* Assume 60 Hz */

   memset(bsregs, 0, 8);
   if (memcmp(bsregs, nsf->bankswitch_info, 8))
   {
      nsf->bankswitched = TRUE;
      log_printf("nsf is bankswitched, load addr $%04X\n", nsf->load_addr);
   }
   else
      nsf->bankswitched = FALSE;

   switch (nsf->ext_sound_type)
   {
   case EXT_SOUND_VRCVI:
      log_printf("VRCVI sound generation detected\n");
      apu_setext(vrcvi_soundprocess);
      break;
   case EXT_SOUND_VRCVII:
      log_printf("VRCVII sound generation detected\n");
      break;
   case EXT_SOUND_FDS:
      log_printf("Famicom Disk System sound detected\n");
      break;
   case EXT_SOUND_NONE:
   default:
      break;
   }
}

/* Load a ROM image into memory */
nsf_t *nsf_load(char *filename)
{
   FILE *fp;
   nsf_t *temp_nsf;

   fp = fopen(filename, "rb");

   /* Didn't find the file?  Maybe the .NSF extension was omitted */
   if (NULL == fp)
   {
      if (NULL == strrchr(filename, '.'))
         strcat(filename, ".nsf");

      fp = fopen(filename, "rb");
      if (NULL == fp)
      {
         log_printf("could not find file '%s'\n", filename);
         return NULL;
      }
   }

   /* bleh */
   temp_nsf = (nsf_t *) malloc(sizeof(nsf_t));

   /* Read in the header */
   fread(temp_nsf, 1, NSF_HEADER_SIZE, fp);

   if (memcmp(temp_nsf->id, NSF_MAGIC, 5))
   {
      fclose(fp);
      log_printf("%s is not an NSF format file\n", filename);
      nsf_free(temp_nsf);
      return NULL;
   }

   /* we're now at position 80h */
   fseek(fp, 0, SEEK_END);
   temp_nsf->length = ftell(fp) - NSF_HEADER_SIZE;

   /* Allocate NSF space, and load it up! */
   /* bleh */
   temp_nsf->data = (uint8 *) malloc(temp_nsf->length);
   if (NULL == temp_nsf->data)
   {
      log_printf("error allocating memory for NSF data\n");
      nsf_free(temp_nsf);
      return NULL;
   }

   /* seek to end of header, read in data */
   fseek(fp, NSF_HEADER_SIZE, SEEK_SET);
   fread(temp_nsf->data, temp_nsf->length, 1, fp);

   fclose(fp);

   /* Set up some variables */
   nsf_setup(temp_nsf);

   log_printf("successfully loaded and initialized %s\n", filename);
/*
   log_printf("load_addr: $%04X\n", temp_nsf->load_addr);
   log_printf("init_addr: $%04X\n", temp_nsf->init_addr);
   log_printf("play_addr: $%04X\n", temp_nsf->play_addr);
*/
   return temp_nsf;
}

/* Free an NSF */
void nsf_free(nsf_t *nsfinfo)
{
   if (nsfinfo->data)
      free(nsfinfo->data);
   if (nsfinfo)
   {
      free(nsfinfo);
      nsfinfo = NULL;
   }

   log_printf("freed all NSF data\n");
}
