/*
** Nofrendo - NES emulator
** (c) 1999 Matthew Conte
**
** License agreement:
**
** This source code may be distributed in unmodified form to any medium,
** and may be used in any *freeware* project.
**
** It may not be sold, in full or in part, as part of any commercial or
** shareware sofware package without the express written permission of 
** the author, Matthew Conte (matt@conte.com).
**
** Modified versions of this code may not be publicly distributed
** without the submission to, and approval of the author (matt@conte.com).
** This includes distribution in source, object, or binary form.
**
** The author claims no responsibility for damage cause by this code; it
** is distributed "as-is"; in other words, use at your own risk.
**
** Any freeware application using this code must credit the author,
** Matthew Conte, in the application itself or the documentation.
**
** Modifications should be reported to matt@conte.com, to ensure the
** highest degree of quality and accuracy is always freely available.
**
**
** nsf.h
**
** NSF loading/saving related defines / prototypes
*/

#ifndef _NSF_H_
#define _NSF_H_

#define  NSF_MAGIC   "NESM\x1A"

#ifdef DJGPP
#define PACKED __attribute__ ((packed))
#else
#define PACKED
#endif

typedef struct nsf_s
{
   uint8 id[5] PACKED;
   uint8 version PACKED;
   uint8 num_songs PACKED;
   uint8 start_song PACKED;
   uint16 load_addr PACKED;
   uint16 init_addr PACKED;
   uint16 play_addr PACKED;
   uint8 song_name[32] PACKED;
   uint8 author_name[32] PACKED;
   uint8 copyright[32] PACKED;
   uint16 ntsc_speed PACKED;
   uint8 bankswitch_info[8] PACKED;
   uint16 pal_speed PACKED;
   uint8 pal_ntsc_bits PACKED;
   uint8 ext_sound_type PACKED;
   uint8 reserved[4] PACKED;

   uint8 *data;
   uint32 length;
   uint32 playback_rate;
   uint8 current_song;
   boolean bankswitched;
} nsf_t;

#define  EXT_SOUND_NONE    0x00
#define  EXT_SOUND_VRCVI   0x01
#define  EXT_SOUND_VRCVII  0x02
#define  EXT_SOUND_FDS     0x04

#define  NSF_HEADER_SIZE   0x80

extern nsf_t *nsf_load(char *filename);
extern void nsf_free(nsf_t *nsf_info);

#endif /* _NSF_H_ */
